|******************************************************************************
|
| project name:    GrayScale-Support for TIGCC
| author:          thomas.nussbaumer@gmx.net
|                  Julien Muchembled (original implementation for UniversalOS)
|                  Kevin@tigcc.ticalc.org (3 plane version - 7/8 grayscales)
|
|
| compatible with HW1/HW2 on all AMS versions up to 2.05
|
|
| $Id: gray.s,v 3.10 2002/04/05 11:34:23 tnussb Exp $
|******************************************************************************

.text
	.xdef Gray3POn,Gray3POff,__gray3P_D_plane,__gray3P_M_plane,__gray3P_L_plane,__gray3P_handle
	.xdef __gray3P_switch_cnt,__gray3P_old_int1_hw1
	.xdef __gray3P_sync_n_count,__gray3P_plane_index
	.xdef __gray3P_dbl_offset,__gray3P_L_plane2,__gray3P_M_plane2,__gray3P_D_plane2

.even
|==============================================================================
| EXPORTED: Gray3POn function (turn grayscales on) - trashes d1/a0/a1
| Parameter: %d0:w = 0 -> 7 grays, non-0 -> 8 grays
|==============================================================================
Gray3POn:
	movem.l  %d2-%d7/%a2-%a6,-(%a7)
	tst.w    %d0
	beq.s    __gray3P_7grays
|Patches for 8 grays:
|__gray3P_phase_reset_hw1
|0x700c = moveq.l #12,%d0
|__gray3P_phase_indir_hw1
|0x3030 000c = move.l 12(%a0,%d0:w),%d0
|__gray3P_phase_reset_hw2
|0x10bc 000c = move.b #12,(%a0)
|__gray3P_phase_indir_hw2
|0x3230 100c = move.l 12(%a0,%d1:w),%d1
        move.w   #0x700c,__gray3P_phase_reset_hw1
        move.w   #0xc,__gray3P_phase_indir_hw1+2
	bra __gray3P_8grays
__gray3P_7grays:
|Patches for 7 grays:
|__gray3P_phase_reset_hw1
|0x700a = moveq.l #10,%d0
|__gray3P_phase_indir_hw1
|0x3030 0000 = move.l 0(%a0,%d0:w),%d0
|__gray3P_phase_reset_hw2
|0x10bc 000a = move.b #10,(%a0)
|__gray3P_phase_indir_hw2
|0x3230 1000 = move.l 0(%a0,%d1:w),%d1
        move.w   #0x700a,__gray3P_phase_reset_hw1
        clr.w    __gray3P_phase_indir_hw1+2
__gray3P_8grays:
	bsr.s    __gray3P_check_hw_version    | evaluate HW version and store it
	lea      (__gray3P_switch_cnt,%pc),%a0| reset plane switch counter to 0
	move.l   #0,(%a0)
	bsr      __gray3P_init_mem            | allocate and initialize memory
	movem.l  (%a7)+,%d2-%d7/%a2-%a6
	move.w   (__gray3P_handle,%pc),%d0
	bne      __gray3P_init_handler        | jump to interrupt handler setup if
	                                      | memory was allocated correctly
	moveq    #0x0,%d0                     | otherwise return 0 (GrayOn failed)
	rts

|==============================================================================
| checks for HW version (VTI is treated as HW1, because port 0x70001D is not
|                        emulated by the VTI and this would cause NEVER switch
|                        planes behaviour if we would use the HW2 support)
|
| returns 0 in d0.w for HW1 and 1 for HW2
|
| IMPORTANT NOTE: This function patches 2 locations in the code of the
|                 grayscale support depending on the HW version. Patching the
|                 locations for both types IS necessary, because if a program
|                 is transfered from one calc to another one the default values
|                 got already overwritten (if program was not archived)
|==============================================================================
__gray3P_check_hw_version:
    |--------------------------------------------------------------------------
    | patches code for HW1 version
    |
    | necessary memory == 2 planes + 8 Bytes == 7688
    |
    | the additionally 8 bytes are necessary for rounding later to a multiple
    | of 8
    |--------------------------------------------------------------------------
__gray3P_patches_for_hw1:
	lea (__gray3P_size_to_allocate,%pc),%a0
	move.w #0x1e08,(%a0)
	move.w #0,(__gray3P_size_to_add-__gray3P_size_to_allocate,%a0)
	rts
|==============================================================================
| Interrupt 1 handler for HW1
|==============================================================================
__gray3P_int1_handler_hw1:
	movem.l  %d0/%a0,-(%a7)
    |--------------------------------------------------------------------------
    | Load skip counter and increment it (count = (count+1)&0x3). Skip any
    | further operation if count is 1, 2 or 3. This means that every 4th call
    | of the INT1 handler is a candidate for a plane switch
    |--------------------------------------------------------------------------
	lea      (__gray3P_skipcount,%pc),%a0
	addq.b   #1,(%a0)
	andi.b   #0x3,(%a0)+            | IMPORTANT: a0 points now to __gray3P_phase!
	bne.s    __gray3P_proceed_old
    |--------------------------------------------------------------------------
    | to evaluate which plane we use counter __gray3P_phase. This counter
    | performs the following counting 8->6->4->2->0(->12)->10->8.
    |--------------------------------------------------------------------------
	move.w   (%a0),%d0
	subq.w   #2,%d0                  | subtract 4 from phase counter
	bcc.s    __gray3P_store          | not negative -> don't reset
__gray3P_phase_reset_hw1:
	moveq    #0x10,%d0               | reset phase counter to 10/12
__gray3P_store:
	move.w   %d0,(%a0)               | store new phase counter value
	lea (__gray3P_7gray_phases),%a0
__gray3P_phase_indir_hw1:
	move.w 0(%a0,%d0:w),%d0
	lea (__gray3P_D_plane,%pc),%a0
    |--------------------------------------------------------------------------
    | doublebuffer extension ... add content of __gray3P_dbl_offset to %d0
    |--------------------------------------------------------------------------
	add.w  (__gray3P_dbl_offset-__gray3P_D_plane,%a0),%d0
	suba.w %d0,%a0
	move.l   (%a0),%d0               | load the address of this plane
	lsr.l    #3,%d0                  | reduce to address / 8
	move.w   %d0,0x600010            | set new plane startaddress
	lea      (__gray3P_switch_cnt,%pc),%a0  | increment switch count
	addq.l   #1,(%a0)
__gray3P_proceed_old:
	movem.l  (%a7)+,%d0/%a0
    |--------------------------------------------------------------------------
    |  JUMP to previous installed interrupt handler
    |--------------------------------------------------------------------------
	.word    0x4ef9                  | "JMP address" opcode
__gray3P_old_int1_hw1:
	.long    0x00000000              | address of old int1 gots stored here
__gray3P_dummy1:                         | NOT used yet (just for alignment)
	.byte    0x00
|------------------------------------------------------------------------------
| __gray3P_skipcount is a one byte counter which performs the following counting:
| 3 -> 0 -> 1 -> 2 -> 3
|------------------------------------------------------------------------------
__gray3P_skipcount:
	.byte    0x03
__gray3P_phase:
	.word    0x08                    | performs: 8->6->4->2->0(->12)->10->8
__gray3P_switch_cnt:
	.long    0x00000000
|==============================================================================
| INTERNAL: allocates memory
|
| modifies: __gray3P_handle
|           __gray3P_used_mem
|           __gray3P_M_plane
|           __gray3P_L_plane
|
| Note: __gray3P_D_plane will not be changed by this function! (will be set by
|                                                        __gray3P_init_handler)
|==============================================================================
__gray3P_init_mem:
	lea      (__gray3P_handle,%pc),%a5          | if __gray3P_handle is not 0
	tst.w    (%a5)                              | we have already allocated
	bne.s    __gray3P_init_return               | memory -> out here
    |--------------------------------------------------------------------------
    | HeapAlloc(HW1=7688 bytes or HW2=11528 bytes)
    |--------------------------------------------------------------------------
	.word    0x4878                       | opcode of "PEA #value"
__gray3P_size_to_allocate:                    | the size gets patched !!
	.word    0x2d08
	jsr      tios__0003
	addq.w   #4,%a7
	move.w   %d0,(%a5)+                   | store handle in handle variable
	beq.s    __gray3P_init_return         | alloc failed (handle=0) -> out here
	clr.w    (%a5)                        | clears __gray3P_dbl_offset
    |--------------------------------------------------------------------------
    | HeapDeref(__gray3P_handle) the dirty way
    |--------------------------------------------------------------------------
	lsl.w    #2,%d0
	move.l   tios__001c+0x1902,%a0
	move.l   0(%a0,%d0:w),%a0
    |--------------------------------------------------------------------------
    | HeapLock(__gray3P_handle) the dirty way
    |--------------------------------------------------------------------------
        bset.b   #7,-2(%a0)
    |--------------------------------------------------------------------------
    | align memory address to next 8-byte boundary and store address in
    | __gray3P_used_mem
    |
    | for HW1: __M_plane gets set to the same address as __gray3P_used_mem
    | for HW2: __M_plane gets set to __gray3P_used_mem + 0xf00
    |--------------------------------------------------------------------------
	move.l   %a0,%d0
	addq.l   #7,%d0
	andi.b   #0xF8,%d0
	lea      (__gray3P_used_mem,%pc),%a0
	move.l   %d0,(%a0)
	.word    0x0680              | opcode of "ADDI.L #value,%a0"
	.word    0x0000
__gray3P_size_to_add:
	.word    0x0F00              | gets patched (HW1:0 HW2:0x0f00)
	move.l   %d0,(__gray3P_M_plane - __gray3P_used_mem,%a0)
	addi.l   #0x0f00,%d0
	move.l   %d0,(__gray3P_L_plane - __gray3P_used_mem,%a0)
__gray3P_init_return:
	rts

|------------------------------------------------------------------------------
| handle to allocated memory used by grayscale
|------------------------------------------------------------------------------
__gray3P_handle:
	.word    0

|------------------------------------------------------------------------------
| DOUBLEBUFFER extension
|------------------------------------------------------------------------------
__gray3P_dbl_offset: | has to be directly AFTER __gray3P_handle!!
	.word    0
__gray3P_L_plane2:
	.long    0x0
__gray3P_M_plane2:
	.long    0x0
__gray3P_D_plane2:
	.long    0x0

|------------------------------------------------------------------------------
| pointer to light plane
| HW1: same as __gray3P_used_mem
| HW2: __gray3P_used_mem + 0xf00
|------------------------------------------------------------------------------
__gray3P_L_plane:
	.long    tios__001c
|------------------------------------------------------------------------------
| pointer to medium plane
| HW1: __gray3P_used_mem + 0xf00
| HW2: __gray3P_used_mem + 0x1e00
|------------------------------------------------------------------------------
__gray3P_M_plane:
	.long    tios__001c
|------------------------------------------------------------------------------
| pointer to dark plane (set by __gray3P_init_handler)
| HW1: tios__001c
| HW2: same as __gray3P_used_mem
|------------------------------------------------------------------------------
__gray3P_D_plane:
	.long    tios__001c
|------------------------------------------------------------------------------
| pointer to allocated memory ALIGNED to 8-byte boundary
|------------------------------------------------------------------------------
__gray3P_used_mem:
	.long    tios__001c
|------------------------------------------------------------------------------
| This variable is very hard to describe. Indeed this isn't one variable,
| but two variables combined in one.
|
| Description will be added later ....
|------------------------------------------------------------------------------
__gray3P_sync_n_count:
	.word    0x0000
|------------------------------------------------------------------------------
| holds the index of the plane which should be drawn next (NOTE: this label
| is never addressed directly, but indirectly from label __gray3P_sync_n_count.
| So don't move it do somewhere else!)
|------------------------------------------------------------------------------
__gray3P_plane_index:
	.word    0x0008
|==============================================================================
| INTERNAL: initialize grayscale handler
|==============================================================================
__gray3P_init_handler:
	lea      (__gray3P_M_plane,%pc),%a0
	move.w   #0x77F,%d1
    |--------------------------------------------------------------------------
    | HW1 specific initializations:
    |
    | (1) "backup" old INT1 handler in __gray3P_old_int1_hw1 (the address part
    |     of a JUMP address instruction at the end of the HW1 int handler)
    | (2) install our own INT1 HW1 handler
    |--------------------------------------------------------------------------
__gray3P_init_hw1_handler:
	move.l   (%a0),%a1
	lea      (__gray3P_int1_handler_hw1,%pc),%a0
	move.l   0x64,(__gray3P_old_int1_hw1 - __gray3P_int1_handler_hw1,%a0)
__gray3P_init_proceed:
	move.l   %a0,0x40064
__gray3P_clr_l_plane:
    |--------------------------------------------------------------------------
    | clear medium and light planes (done for both HW types)
    |--------------------------------------------------------------------------
	clr.l    (%a1)+
	dbf      %d1, __gray3P_clr_l_plane
    |--------------------------------------------------------------------------
    | PortSet(__D_plane,239,127)
    |--------------------------------------------------------------------------
	move.l   #0xEF007F,-(%a7)
	move.l   (__gray3P_D_plane,%pc),-(%a7)
        jsr      tios__0015
	addq.w   #8,%a7
__gray3P_ok:
	lea (__gray3P_L_plane,%pc),%a0
	lea (__gray3P_L_plane2,%pc),%a1
	move.l (%a0)+,(%a1)+        | copy __L_plane to __L_plane2
	move.l (%a0)+,(%a1)+        | copy __M_plane to __M_plane2
	move.l (%a0)+,(%a1)+        | copy __D_plane to __D_plane2
	moveq    #0x1,%d0
	rts
|==============================================================================
|  EXPORTED: Gray3POff function (turn grayscales off)
|            NOTE: ALWAYS returns 1 !!
|==============================================================================
Gray3POff:
	movem.l  %a2-%a3,-(%a7)
	lea      (__gray3P_handle,%pc),%a3
	tst.w    (%a3)
	beq      __gray3P_off_out                   | no handle? -> nothing to do
    |--------------------------------------------------------------------------
    | cleanup for HW1 calcs (on HW1 tios__001c is used as darkplane. We haven't to
    | set it)
    |--------------------------------------------------------------------------
__gray3P_hw1_cleanup:
        move.l   #tios__001c,%d0
        lsr.l    #3,%d0
	move.w   %d0,0x600010                        | restore used plane to tios__001c
	move.l   (__gray3P_old_int1_hw1,%pc),0x40064 | restore old INT1 handler

__gray3P_continue_cleanup:
	lea      (__gray3P_L_plane,%pc),%a0  | restore plane pointers to tios__001c for sure
	move.l   #tios__001c,(%a0)+
	move.l   #tios__001c,(%a0)+
	move.l   #tios__001c,(%a0)+
	move.l   #tios__001c,(%a0)
    |--------------------------------------------------------------------------
    | HeapFree(__gray3P_handle)
    |--------------------------------------------------------------------------
	move.w   (%a3),-(%a7)
	jsr      tios__0002
	addq.l   #2,%a7
    |--------------------------------------------------------------------------
    | PortRestore()
    |--------------------------------------------------------------------------
	jsr      tios__0016
	clr.l    (%a3)                     | 0->handle AND(!!) 0->__gray3P_dbl_offset
	lea      (__gray3P_sync_n_count,%pc),%a0
	clr.l    (%a0)
__gray3P_off_out:
	movem.l  (%a7)+,%a2-%a3
	jbra     __gray3P_ok

__gray3P_7gray_phases:
	.word	8
	.word	0
	.word	4
	.word	0
	.word	4
	.word	0

__gray3P_8gray_phases:
	.word	4
	.word	0
	.word	8
	.word	0
	.word	0
	.word	4
	.word	0

| #############################################################################
|  Revision History
| #############################################################################
|
| Revision 3.10-3P-4-F1 2003-09-26 Kevin Kofler
| Use .text instead of .data.
|
| Revision 3.10-3P-3-F1 2003-02-10 Kevin Kofler
| Ported to Fargo II
|
| Revision 3.10-3P-3 2003-01-14 Kevin Kofler
| Fixed misleading comment in the code automodifying the HW2 phase reset
|
| Revision 3.10-3P-2 2002-09-06 Kevin Kofler
| Fixed typo in switch counting code
|
| Revision 3.10-3P 2002-07-29 Kevin Kofler
| Grayscale routine adapted to handle 7/8 grayscales
|
| $Log: gray.s,v $
| Revision 3.10 2002/04/05 11:34:23  tnussb
| (1) Resets now __D_plane2,__L_plane2 and __gray_dbl_offset at end of GrayOn()
|     to make sure nothing happens if doublebuffer macros get called without
|     setting a buffer first. Nevertheless NO program should call one of the
|     doublebuffer macros without setting a buffer previously.
| (2) Some further size optimizations. Now the size is exactly as long as
|     without before the integration of doublebuffering. Quite smart, isn't it?
|     More functionality, some minor drawbacks fixed and no increase in size ...
| (3) Changed return value of GrayOff() function to void.
|
| Revision 3.9  2002/04/04 18:50:36  tnussb
| first working version of internal doublebuffer support (not checked on
| real calcs yet, but on the VTI)
|
| Revision 3.8  2002/04/04 16:39:05  tnussb
| Debug version checked in previously. This one is the correct one.
|
| Revision 3.7  2002/04/04 16:35:20  tnussb
| (1) documentation of HW2 interrupt handler heavily extended
| (2) HW2: plane switch counting fixed (was incremented too often previously)
| (3) global ALWAYS_HW2_TESTING added which can be used to force the use of
|     HW2 grayscales. Now the HW2 grayscales can be tested on the VTI if
|     this global is defined. Of course this will flicker extremely, because
|     the VTI doesn't simulate port 70001D, but its still better than no
|     testing possibility.
| (4) don't trashes %sr on HW2 anymore (restores previous setting)
| (5) This version was tested on a real HW1 and a real HW2 calc. It works.
|     (Thanx to Sebastian Reichelt and stoopid guy)
|
| Revision 3.6  2002/04/04 11:58:19  tnussb
| (1) size optimizations
| (2) unnecessary cleanup removed from GrayOff()
| (3) interrupt handler for HW1 rewritten (uses now plane pointer directly)
| (4) "heavily" documented
|
| Revision 3.5  2002/04/04 11:54:39  tnussb
| (1) exports __gray_old_int1_hw1 and __gray_old_int1_hw2. this way it is
|     possible to modify them after grayscales are turned on
| (2) comments changed to GNU Assembler style
| [NOTE: CVS time and date doesn't fit to real implementation data]
|
| Revision 3.0  2002/04/04 11:50:56  tnussb
| grayscale support used for TIGCC up to version v0.93
| [NOTE: CVS time and date doesn't fit to real implementation data]
|
